/* Copyright (C) 2024 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer] */

#pragma once

#include <type_traits>

/**
 * @file AlgoTraits.h
 * @brief Type traits for online algorithms
 */

namespace cbm::algo::algo_traits
{

  namespace detail
  {
    template<typename...>
    struct ResultOf {
    };

    template<typename R, typename Algo, typename... Args>
    struct ResultOf<R (Algo::*)(Args...) const> {
      using type = R;
    };

    template<typename R, typename Algo, typename... Args>
    struct ResultOf<R (Algo::*)(Args...)> {
      using type = R;
    };

    template<typename Algo>
    struct ResultOf<Algo> : ResultOf<decltype(&Algo::operator())> {
    };

  }  // namespace detail

  /**
   * @brief Type alias for the return type produced by an algorithm when invoked via callable-operator
   */
  template<typename Algo>
  using ResultOf_t = typename detail::ResultOf<Algo>::type;

  // Currently assume algorithms return std::pair<R, M>
  // where R is the output and M is the monitoring data

  /**
   * @brief Type alias for the output type produced by an algorithm
   */
  template<typename Algo>
  using Output_t = typename ResultOf_t<Algo>::first_type;

  /**
   * @brief Type alias for the monitoring type produced by an algorithm
   */
  template<typename Algo>
  using Monitor_t = typename ResultOf_t<Algo>::second_type;

}  // namespace cbm::algo::algo_traits
